﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Threading;

namespace ir_kit
{
    public class BrainBandParser
    {
        int MAX_PACKET_LENGTH = 170;
        int EEG_POWER_BANDS = 8;
        int RAW_DATA_LENGTH = 800;

        SerialPort port;
        int latestByte, lastByte, lastPacket;
        bool inPacket;
        byte[] packetData;
        int rawIndex;
        int[] rawData, rawViewData;

        int signalQuality, attention, meditation, blink;
        int[] eegPower;
        int minPower = (int)Math.Pow(2, 31), maxPower = (int)-Math.Pow(2, 31);

        Queue<byte> fifo = new Queue<byte>();

        public int LastPacket { get { return lastPacket; } }
        public int RawIndex { get { return rawIndex; } }
        public int[] RawData
        {
            get
            {
                for(int i=0;i<RAW_DATA_LENGTH;i++)
                {
                    int j = rawIndex - i;
                    if (j < 0) j += RAW_DATA_LENGTH;
                    rawViewData[j] = rawData[i];
                }
                return rawViewData;
            }
        }

        public int SignalQuality { get { return signalQuality; } }
        public int Attention { get {return attention;} }
        public int Meditation { get {return meditation; } }
        public int Delta { get { return eegPower[0]; } }
        public int Theta { get { return eegPower[1]; } }
        public int LowAlpha { get { return eegPower[2]; } }
        public int HighAlpha { get { return eegPower[3]; } }
        public int LowBeta { get { return eegPower[4]; } }
        public int HighBeta { get { return eegPower[5]; } }
        public int LowGamma { get { return eegPower[6]; } }
        public int MidGamma { get { return eegPower[7]; } }

        public int MinPower { get { return minPower; } }
        public int MaxPower { get { return maxPower; } }

        override public String ToString()
        {
            String ret = "";
            ret += "Signal Quality: " + signalQuality + "\n";
            ret += "Attention: " + attention + "\n";
            ret += "Meditation: " + meditation + "\n";
            ret += "Blink: " + blink + "\n";
            ret += "Delta: " + eegPower[0] + "\n";
            ret += "Theta: " + eegPower[1] + "\n";
            ret += "Low Alpha: " + eegPower[2] + "\n";
            ret += "High Alpha: " + eegPower[3] + "\n";
            ret += "Low Beta: " + eegPower[4] + "\n";
            ret += "High Beta: " + eegPower[5] + "\n";
            ret += "Low Gamma: " + eegPower[6] + "\n";
            ret += "Mid Gamma: " + eegPower[7];

            return ret;
        }

        public BrainBandParser(String p)
            : this(p, 57600)
        {
        }

        public BrainBandParser(String p, int b)
        {
            port = new SerialPort(p, b, Parity.None, 8, StopBits.One);
            if (port != null)
            {
                port.Handshake = Handshake.None;
                port.ReadBufferSize = 1024 * 1024;
                port.Open();
                if (!port.IsOpen)
                {
                    throw new Exception("SerialPort can't open.");
                }
            }
            else
            {
                throw new Exception("SerialPort can't create.");
            }

            latestByte = lastByte = lastPacket = 0;
            inPacket = false;
            packetData = new byte[MAX_PACKET_LENGTH];
            eegPower = new int[EEG_POWER_BANDS];
            rawIndex = 0;
            rawData = new int[RAW_DATA_LENGTH];
            rawViewData = new int[RAW_DATA_LENGTH];
        }

        ~BrainBandParser()
        {
            close();
        }

        public void close()
        {
            if (port != null && port.IsOpen)
                port.Close();
        }

        bool parsePacket(int len)
        {
            int extCodeLevel = 0, vlen = 0;
            for (int i = 0; i < len; i++)
            {
                lastPacket = packetData[i];
                switch (lastPacket)
                {
                    case 0x02: signalQuality = packetData[++i]; break;
                    case 0x04: attention = packetData[++i]; break;
                    case 0x05: meditation = packetData[++i]; break;
                    case 0x16: blink = packetData[++i]; break;
                    case 0x55: extCodeLevel++; break;
                    case 0x80:
                        vlen = packetData[++i];
                        int v = (packetData[++i] << 8) | packetData[++i];
                        if (v >= 32768) v -= 65536;
                        rawData[rawIndex++] = v;
                        if (rawIndex >= RAW_DATA_LENGTH)
                        {
                            rawIndex = 0;
                        }
                        break;
                    case 0x83:
                        vlen = packetData[++i];
                        if (vlen > EEG_POWER_BANDS * 3)
                        {
                            Console.WriteLine("ERROR: VLENGTH TOO LARGE " + vlen);
                            i = len;
                        }
                        else
                        {
                            for (int j = 0; j < EEG_POWER_BANDS; j++)
                            {
                                eegPower[j] = (packetData[++i] << 16) | (packetData[++i] << 8) | packetData[++i];
                                if (minPower > eegPower[j]) minPower = eegPower[j];
                                if (maxPower < eegPower[j]) maxPower = eegPower[j];
                            }
                        }
                        break;
                    default:
                        return false;
                }
            }
            return true;
        }

        int getpayload(int len)
        {
            int chksum = 0;
            for (int i = 0; i < len; i++)
            {
                chksum += (packetData[i] = read_byte());
            }
            chksum &= 0xff;
            chksum = ~chksum & 0xff;

            return chksum;
        }

        bool inpacket(int len)
        {
            if (len > MAX_PACKET_LENGTH)
            {
                Console.WriteLine("ERROR: PLENGTH TOO LARGE");
                return false;
            }
            int chksum = getpayload(len);
            if (chksum != read_byte()) return false;

            return true;
        }

        private int read_data()
        {
            if (port.BytesToRead == 0) return 0;
            byte[] buf = new byte[port.BytesToRead];
            port.Read(buf, 0, buf.Length);

            foreach (byte b in buf)
                fifo.Enqueue(b);

            return buf.Length;
        }

        private byte read_byte()
        {
            if (fifo.Count == 0)
            {
                int n = 0;
                while (read_data() == 0)
                {
                    if (n++ > port.ReadTimeout)
                        throw new Exception("serial read timeout");
                    Thread.Sleep(1);
                }
            }
            return fifo.Dequeue();
        }

        public bool update()
        {
            bool ret = false;
            int plen;

            while (port.BytesToRead > 0)
            {
                port.ReadTimeout = 100;    // timeout (msec)
                latestByte = read_byte();

                if (inPacket)
                {
                    plen = latestByte;
                    if (plen != 170)    // 0xaa
                    {
                        if (inpacket(plen))
                        {
                            if (parsePacket(plen))
                            {
                                ret = true;
                            }
                            else
                            {
                                Console.WriteLine("ERROR: COULD NOT PARSE");
                            }
                        }
                        else
                        {
                            Console.WriteLine("ERROR: CHECKSUM");
                        }
                        lastByte = latestByte = 0;
                        return true;
                    }
                    inPacket = false;
                }
                else if ((latestByte == 0xaa) && (lastByte == 0xaa))
                {
                    inPacket = true;
                }
                lastByte = latestByte;
            }
            return ret;
        }
    }
}
